{*******************************************************************************
  Copyright (C) Michael W. Vogel m-w-vogel@gmx.de

  The contents of this file are subject to the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 3.0 of the License,
  or (at your option) any later version.
  You may obtain a copy of the LGPL at http://www.gnu.org/copyleft/.

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
  specific language governing rights and limitations under the License.

  Created 02.09.2015
*******************************************************************************}

unit Defaces;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Dialogs, zipper, DCPrijndael,
  DCPSha256, ArchiveParser, Progresses, DefacerStrings;

type
  TByteArray = array of Byte;
  EDefaceError = class(Exception);
  TDefaceState = (dsEmpty, dsFilesAvial, dsZipped, dsEncrypted, dsCoded);

  { TCodeData }

  TCodeData = class(TObject)
  private
    FData: TByteArray;                                      //Zur Verschlüsselung genutzte Daten
  public
    procedure LoadFromFile(const aFileName: String);        //Lädt Data aus File
    property Data: TByteArray read FData write FData;       //Zur Verschlüsselung genutzte Daten
  end;

  { TCustomDefacer }

  TCustomDefacer = class(TObject)
  private
    FFileName: String;                                      //Filename des Archives
    FOnArchiveFileNamesRead: TNotifyEvent;                  //Wenn Filenamen aus Archiv extrahiert wurden
    FState: TDefaceState;                                   //In welchem Zustand befinden sich die Streams?!
    FEncStream: TMemoryStream;                              //Stream zur Verschlüsselung
    FTmpStream: TMemoryStream;                              //Stream zur Verschlüsselung
    FCodeData: TCodeData;                                   //Zur Ver- und Entschlüsselung genutzte Data
    FCodeFileList: TStringList;                             //Welche Files sollen für Verschlüsselung genutzt werden?!
    FOnDefaceProgress: TProgressEvent;                      //Progress bei Ver-/Entschlüsselung
    FPackFileList: TStringList;                             //Welche Files sollen ver- / entschlüsselt werden - getrennt nach DiskFileName und ArchiveFileName
    FPassWord: String;                                      //Genutztes Passwort
    FOutputDir: String;                                     //Wohin ein Archiv entpackt werden soll
    FUseStream: Boolean;                                    //Soll in ein Stream gespeichert werden?!
    procedure SetPassWord(AValue: String);                  //Setter für Passwort
    function ZipFiles: Boolean;                             //Packt Files
    function UnZipFiles: Boolean;                           //Entpackt Files
    function ReadZipFileNames: Boolean;                     //Können aus Gepackten Daten Filenamen extrahiert werden?!
    procedure ZipProgress({%H-}Sender: TObject;             //Progress beim Packen
      Const AFileName : String);
    procedure GetZipStream({%H-}Sender: TObject;            //Verwendung des Streams statt einer Datei zum Entpacken
      var AStream: TStream);
    function EncryptZip: Boolean;                           //Verschlüsselt den zuvor gezippten Stream
    function DecryptZip: Boolean;                           //Entschlüssel und erstelle ZipFile
    procedure EncodeOneFile(CodeFileName: String);          //Verschlüsselt erstelltes verschlüsselte Datei mit Datei
    procedure DecodeOneFile(CodeFileName: String);          //Entschlüsselt erstelltes verschlüsselte Datei mit Datei
    procedure EncodeFile;                                   //Verschlüsselt erstelltes verschlüsselte Datei mit Dateien
    procedure DecodeFile;                                   //Entschlüsselt erstelltes verschlüsselte Datei mit Dateien
    procedure SaveDefaceFile;                               //Speichert das DefaceFile
  protected
    property OnDefaceProgress: TProgressEvent               //Progress bei Ver-/Entschlüsselung
      read FOnDefaceProgress write FOnDefaceProgress;
    procedure ProgressStart; virtual;                       //Summen, wie lange der jeweilige Teil dauert berechnen
    procedure InDeface; virtual;                            //Verschlüsselung
    procedure OutDeface; virtual;                           //Entschlüsselung
    procedure LoadFromFile(aFileName: String); virtual;     //Laden vom File
    procedure LoadFromStream(Stream: TStream); virtual;     //Laden vom Stream
    procedure SaveToFile(aFileName: String); virtual;       //Speichern in Datei
    procedure SaveToStream(Stream: TStream); virtual;       //Speichern in Stream
    procedure SaveToDirectory(Dir: String); virtual;        //Speichern der Dateien in Verzeichnis
    property CodeData: TCodeData read FCodeData;            //Zur Ver- und Entschlüsselung genutzte Data
    property CodeFileList: TStringList read FCodeFileList;  //Welche Files sollen für Verschlüsselung genutzt werden?!
    property PackFileList: TStringList                      //Welche Files sollen ver- / entschlüsselt werden - FileNamen müssem geparst werden
      read FPackFileList;
    property PassWord: String read FPassWord                //Genutztes Passwort
      write SetPassWord;
    property OnArchiveFileNamesRead: TNotifyEvent           //Wenn Filenamen aus Archiv extrahiert wurden
      read FOnArchiveFileNamesRead write FOnArchiveFileNamesRead;
    property FileName: String read FFileName                //Name des gepackten Archives
      write FFilename;
    property OutPutDir: String read FOutputDir              //Wohin ein Archiv entpackt werden soll
      write FOutPutDir;
  public
    constructor Create;                                     //Constructor
    destructor Destroy; override;                           //Destructor
    procedure Clear;                                        //Clear
    procedure CodeFileAdd(const aFileName: String);         //Fügt ein File zur CodeFileList hinzu
    procedure CodeFileDel(aIndex: Integer);                 //Löscht File an Index...
  end;

  { TInDefacer }

  TInDefacer = class(TCustomDefacer)
  public
    procedure InDeface; override;
    procedure ProgressStart; override;
    procedure SaveToFile(aFileName: String); override;
    procedure SaveToStream(Stream: TStream); override;
  published
    property FileName;
    property OnDefaceProgress;
    property CodeFileList;
    property PackFileList;
    property PassWord;
  end;

  { TOutDefacer }

  TOutDefacer = class(TCustomDefacer)
  public
    procedure ProgressStart; override;
    procedure OutDeface; override;
    procedure LoadFromFile(aFileName: String); override;    //Laden vom File
    procedure LoadFromStream(Stream: TStream); override;    //Laden vom Stream
    procedure SaveToDirectory(Dir: String); override;       //Speichern der Dateien in Verzeichnis
  published
    property FileName;
    property OutPutDir;
    property OnDefaceProgress;
    property OnArchiveFileNamesRead;
    property CodeFileList;
    property PackFileList;
    property PassWord;
  end;

implementation

{$PUSH}                                                     // Compiler-Einstellungen speichern
{$RANGECHECKS OFF}

function EncodeAdd(const Orig, Code: Byte): Byte; inline;
begin
  Result := Orig + Code;
end;
function EncodeSub(const Orig, Code: Byte): Byte; inline;
begin
  Result := Orig - Code;
end;
function EncodeShl(const Orig, Code: Byte): Byte; inline;
var
  w: Word;
begin
  w := Orig;
  w := w shl (Code mod 8);
  Result := w + w shr 8;
end;
function EncodeShr(const Orig, Code: Byte): Byte; inline;
var
  w: Word;
begin
  w := Orig shl 8;
  w := w shr (Code mod 8);
  Result := w + w shr 8;
end;
function EncodeXor(const Orig, Code: Byte): Byte; inline;
begin
  Result := Orig xor Code;
end;
function DecodeAdd(const Orig, Code: Byte): Byte; inline;
begin
  Result := EncodeSub(Orig, Code);
end;
function DecodeSub(const Orig, Code: Byte): Byte; inline;
begin
  Result := EncodeAdd(Orig, Code);
end;
function DecodeShl(const Orig, Code: Byte): Byte; inline;
begin
  Result := EncodeShr(Orig, Code);
end;
function DecodeShr(const Orig, Code: Byte): Byte; inline;
begin
  Result := EncodeShl(Orig, Code);
end;
function DecodeXor(const Orig, Code: Byte): Byte; inline;
begin
  Result := EncodeXor(Orig, Code);
end;

{$POP}                                                      // gespeicherte Einstellungen wiederherstellen

{ TCodeData }

procedure TCodeData.LoadFromFile(const aFileName: String);  //Lädt Data aus File
var
  FS: TFileStream;
begin
  try
    FS := TFileStream.Create(aFileName, fmOpenRead);
    SetLength(FData, FS.Size);
    FS.Read(FData[0], FS.Size);
  finally
    FreeAndNil(FS);
  end;
end;

{ TCustomDefacer }

procedure TCustomDefacer.SetPassWord(AValue: String);       //Setter für Passwort
begin
  if FPassWord = AValue then Exit;
  FPassWord := AValue;
end;

function TCustomDefacer.ZipFiles: Boolean;                  //Packt Files

var
  Zipper: TZipper;

  procedure DefineZipEntries;
  var
    FileNr: Integer;
  begin
    Progress.StartNext;
    Progress.DefineRealSteps(PackFileList.Count);
    for FileNr := 0 to PackFileList.Count - 1 do begin
      if Progress.IsStateCallingSetStep(1, FileNr) then
        Progress.State(ArchiveFileName(PackFileList[FileNr]));
      Zipper.Entries.AddFileEntry(
        DiskFileName(PackFileList[FileNr]),
        ArchiveFileName(PackFileList[FileNr]));
    end;
  end;

begin
  Result := False;
  Zipper := TZipper.Create;
  Zipper.OnStartFile := @ZipProgress;
  try
    DefineZipEntries;
    if Zipper.Entries.Count > 0 then begin
      Progress.StartNext;
      Progress.DefineRealSteps(Zipper.Entries.Count);
      Zipper.SaveToStream(FTmpStream);
      FState := dsZipped;
      Result := True;
    end;
  except
    on e: Exception do
      ShowMessage(e.Message);
  end;
  FreeAndNil(Zipper);
end;

function TCustomDefacer.UnZipFiles: Boolean;                //Entpackt Files
var
  UnZipper: TUnZipper;
begin
  Result := False;
  if (FTmpStream.Size = 0) or (FState <> dsZipped) then
    raise EDefaceError.Create(SErrNoZippedData);

  Progress.StartNext;
  Progress.DefineRealSteps(PackFileList.Count);
  UnZipper := TUnZipper.Create;
  UnZipper.OnOpenInputStream := @GetZipStream;
  UnZipper.OnStartFile := @ZipProgress;
  try
    UnZipper.OutputPath := FOutputDir;
    UnZipper.UnZipAllFiles;
    Result := True;
  except
    on e: Exception do
      ShowMessage(e.Message);
  end;
  FreeAndNil(UnZipper);
end;

function TCustomDefacer.ReadZipFileNames: Boolean;          //Können aus Gepackten Daten Filenamen extrahiert werden?!
var
  UnZipper: TUnZipper;

  procedure FillPackFileList;
  var
    i: Integer;
  begin
    PackFileList.Clear;
    for i := 0 to UnZipper.Entries.Count - 1 do begin
      if Progress.IsStateCallingAddStep then
        Progress.State(UnZipper.Entries[i].ArchiveFileName);
      if UnZipper.Entries[i].IsDirectory then
        PackFileList.Add(PackFileName(True, UnZipper.Entries[i].DiskFileName, UnZipper.Entries[i].ArchiveFileName))
      else
        PackFileList.Add(PackFileName(False, UnZipper.Entries[i].DiskFileName, UnZipper.Entries[i].ArchiveFileName));
    end;
  end;

begin
  if FState <> dsZipped then
    raise EDefaceError.Create(SErrNoZippedData);

  Result := False;
  UnZipper := TUnZipper.Create;
  UnZipper.OnOpenInputStream := @GetZipStream;
  try
    UnZipper.Examine;
    if UnZipper.Entries.Count > 0 then begin
      Progress.StartNext;
      Progress.DefineRealSteps(UnZipper.Entries.Count);
      FillPackFileList;
      if Assigned(FOnArchiveFileNamesRead) then
        FOnArchiveFileNamesRead(Self);
      Result := True;
    end;
  except
    on e: Exception do
      ShowMessage(e.Message);
  end;
  FreeAndNil(UnZipper);
end;

procedure TCustomDefacer.ZipProgress(Sender: TObject;       //Progress beim Packen
  const AFileName: String);
begin
  if Progress.IsStateCallingAddStep then
    Progress.State(AFileName);
end;

procedure TCustomDefacer.GetZipStream(Sender: TObject;      //Verwendung des Streams statt einer Datei zum Entpacken
  var AStream: TStream);
begin
  AStream := TMemoryStream.Create;
  FTmpStream.Position := 0;
  AStream.CopyFrom(FTmpStream, FTmpStream.Size);
end;

function TCustomDefacer.EncryptZip: Boolean;                //Verschlüsselt den zuvor gezippten Stream
var
  LCrypt: TDCP_rijndael;
begin
  Result := False;
  FTmpStream.Position := 0;
  FEncStream.Position := 0;
  LCrypt := TDCP_rijndael.Create(nil);
  try
    LCrypt.InitStr(Password, TDCP_sha256);
    Progress.StartNext;
    Progress.State('');
    LCrypt.EncryptStream(FTmpStream, FEncStream, FTmpStream.Size);
    Progress.StartNext;
    Progress.State('');
    LCrypt.Burn;
    FState := dsEncrypted;
    Result := True;
  except
    on e: Exception do
      ShowMessage(e.Message);
  end;
  FreeAndNil(LCrypt);
end;

function TCustomDefacer.DecryptZip: Boolean;                //Entschlüssel und erstelle ZipFile
var
  LCrypt: TDCP_rijndael;
begin
  Result := False;
  FTmpStream.Position := 0;
  FEncStream.Position := 0;
  LCrypt := TDCP_rijndael.Create(nil);
  try
    LCrypt.InitStr(Password, TDCP_sha256);
    Progress.StartNext;
    Progress.State('');
    LCrypt.DecryptStream(FEncStream, FTmpStream, FEncStream.Size);
    Progress.StartNext;
    Progress.State('');
    LCrypt.Burn;
    FState := dsZipped;
    Result := True;
  except
    on e: Exception do
      ShowMessage(e.Message);
  end;
  FEncStream.Size := 0;
  FreeAndNil(LCrypt);
end;

procedure TCustomDefacer.EncodeOneFile(                     //Verschlüsselt erstelltes verschlüsselte Datei mit Datei
  CodeFileName: String);
var
  CodData: TByteArray;
  PosPass, PosCode: Integer;
  PosStream: Integer;
  bTmp, bCode, bEnc, bMove, bPass: Byte;

  procedure ReadCodeData;
  var
    CodFileStream: TFileStream;
  begin
    CodFileStream := TFileStream.Create(CodeFilename, fmOpenRead);
    SetLength(CodData, CodFileStream.Size);
    CodFileStream.ReadBuffer(CodData[0], CodFileStream.Size);
    FreeAndNil(CodFileStream);
  end;

begin
  FEncStream.Position := 0;
  FTmpStream.Position := 0;
  ReadCodeData;

  PosCode := 0;
  PosPass := 1;

  {$PUSH}                                                   // Compiler-Einstellungen speichern
  {$RANGECHECKS OFF}

  for PosStream := 0 to FEncStream.Size - 1 do begin
    bTmp  := FEncStream.ReadByte;
    bCode := CodData[PosCode];
    bPass := Byte(PassWord[PosPass]);
    bCode := bCode + bPass;

    case bCode mod 21 of
      0 : bEnc := EncodeAdd(bTmp, bCode);
      1 : bEnc := EncodeSub(bTmp, bCode);
      2 : bEnc := EncodeShl(bTmp, bCode);
      3 : bEnc := EncodeShr(bTmp, bCode);
      4 : bEnc := EncodeXor(bTmp, bCode);
      5 : bEnc := EncodeAdd(EncodeShl(bTmp, bCode), bCode);
      6 : bEnc := EncodeAdd(EncodeShr(bTmp, bCode), bCode);
      7 : bEnc := EncodeAdd(EncodeXor(bTmp, bCode), bCode);
      8 : bEnc := EncodeSub(EncodeShl(bTmp, bCode), bCode);
      9 : bEnc := EncodeSub(EncodeShr(bTmp, bCode), bCode);
      10: bEnc := EncodeSub(EncodeXor(bTmp, bCode), bCode);
      11: bEnc := EncodeShl(EncodeAdd(bTmp, bCode), bCode);
      12: bEnc := EncodeShl(EncodeSub(bTmp, bCode), bCode);
      13: bEnc := EncodeShl(EncodeXor(bTmp, bCode), bCode);
      14: bEnc := EncodeShr(EncodeAdd(bTmp, bCode), bCode);
      15: bEnc := EncodeShr(EncodeSub(bTmp, bCode), bCode);
      16: bEnc := EncodeShr(EncodeXor(bTmp, bCode), bCode);
      17: bEnc := EncodeXor(EncodeAdd(bTmp, bCode), bCode);
      18: bEnc := EncodeXor(EncodeSub(bTmp, bCode), bCode);
      19: bEnc := EncodeXor(EncodeShl(bTmp, bCode), bCode);
      20: bEnc := EncodeXor(EncodeShr(bTmp, bCode), bCode);
    end;

    FTmpStream.WriteByte(bEnc);

    bMove := bEnc + bCode;
    PosCode := (PosCode + bMove) mod Length(CodData);

    Inc(PosCode);
    Inc(PosPass);
    if PosPass > Length(PassWord) then PosPass := 1;

    if (PosStream mod 1000000 = 0) and
      Progress.IsStateCallingSetStep(PosStream) then
      Progress.State('');
  end;

  {$POP}                                                    // gespeicherte Einstellungen wiederherstellen

  FEncStream.Position := 0;
  FTmpStream.Position := 0;
  FTmpStream.SaveToStream(FEncStream);
end;

procedure TCustomDefacer.DecodeOneFile(                     //Verschlüsselt erstelltes verschlüsselte Datei mit Datei
  CodeFileName: String);
var
  CodData: TByteArray;
  PosPass, PosCode: Integer;
  PosStream: Integer;
  bTmp, bCode, bDec, bMove, bPass: Byte;

  procedure ReadCodeData;
  var
    CodFileStream: TFileStream;
  begin
    CodFileStream := TFileStream.Create(CodeFilename, fmOpenRead);
    SetLength(CodData, CodFileStream.Size);
    CodFileStream.ReadBuffer(CodData[0], CodFileStream.Size);
    FreeAndNil(CodFileStream);
  end;

begin
  FEncStream.Position := 0;
  FTmpStream.Position := 0;
  ReadCodeData;

  PosCode := 0;
  PosPass := 1;

  {$PUSH}                                                   // Compiler-Einstellungen speichern
  {$RANGECHECKS OFF}

  for PosStream := 0 to FEncStream.Size - 1 do begin
    bTmp  := FEncStream.ReadByte;
    bCode := CodData[PosCode];
    bPass := Byte(PassWord[PosPass]);
    bCode := bCode + bPass;

    case bCode mod 21 of
      0 : bDec := DecodeAdd(bTmp, bCode);
      1 : bDec := DecodeSub(bTmp, bCode);
      2 : bDec := DecodeShl(bTmp, bCode);
      3 : bDec := DecodeShr(bTmp, bCode);
      4 : bDec := DecodeXor(bTmp, bCode);
      5 : bDec := DecodeShl(DecodeAdd(bTmp, bCode), bCode);
      6 : bDec := DecodeShr(DecodeAdd(bTmp, bCode), bCode);
      7 : bDec := DecodeXor(DecodeAdd(bTmp, bCode), bCode);
      8 : bDec := DecodeShl(DecodeSub(bTmp, bCode), bCode);
      9 : bDec := DecodeShr(DecodeSub(bTmp, bCode), bCode);
      10: bDec := DecodeXor(DecodeSub(bTmp, bCode), bCode);
      11: bDec := DecodeAdd(DecodeShl(bTmp, bCode), bCode);
      12: bDec := DecodeSub(DecodeShl(bTmp, bCode), bCode);
      13: bDec := DecodeXor(DecodeShl(bTmp, bCode), bCode);
      14: bDec := DecodeAdd(DecodeShr(bTmp, bCode), bCode);
      15: bDec := DecodeSub(DecodeShr(bTmp, bCode), bCode);
      16: bDec := DecodeXor(DecodeShr(bTmp, bCode), bCode);
      17: bDec := DecodeAdd(DecodeXor(bTmp, bCode), bCode);
      18: bDec := DecodeSub(DecodeXor(bTmp, bCode), bCode);
      19: bDec := DecodeShl(DecodeXor(bTmp, bCode), bCode);
      20: bDec := DecodeShr(DecodeXor(bTmp, bCode), bCode);
    end;

    FTmpStream.WriteByte(bDec);

    bMove := bTmp + bCode;
    PosCode := (PosCode + bMove) mod Length(CodData);

    Inc(PosCode);
    Inc(PosPass);
    if PosPass > Length(PassWord) then PosPass := 1;

    if (PosStream mod 1000000 = 0) and Progress.IsStateCallingSetStep(PosStream) then
      Progress.State('');
  end;

  {$POP}                                                    // gespeicherte Einstellungen wiederherstellen

  FEncStream.Position := 0;
  FTmpStream.Position := 0;
  FTmpStream.SaveToStream(FEncStream);
end;

procedure TCustomDefacer.EncodeFile;                        //Verschlüsselt erstelltes verschlüsselte Datei mit Dateien
var
  i: Integer;
begin
  for i := 0 to FCodeFileList.Count - 1 do begin
    Progress.StartNext;
    Progress.DefineRealSteps(FEncStream.Size);
    EncodeOneFile(FCodeFileList[i]);
  end;
  FState := dsCoded;
end;

procedure TCustomDefacer.DecodeFile;                        //Entschlüsselt erstelltes verschlüsselte Datei mit Dateien
var
  i: Integer;
begin
  for i := FCodeFileList.Count - 1 downto 0 do begin
    Progress.StartNext;
    Progress.DefineRealSteps(FEncStream.Size);
    DecodeOneFile(FCodeFileList[i]);
  end;
  FState := dsEncrypted;
end;

procedure TCustomDefacer.SaveDefaceFile;                    //Speichert das DefaceFile
begin
  Progress.StartNext;
  Progress.State(FFileName);
  FEncStream.SaveToFile(FFileName);
end;

procedure TCustomDefacer.ProgressStart;                     //Summen, wie lange der jeweilige Teil dauert berechnen
begin
  Progress.Clear;
  Progress.Visible := True;
end;


procedure TCustomDefacer.InDeface;                          //Verschlüsselung
begin
  try
    if PackFileList.Count < 1 then
      raise EDefaceError.Create(SErrNoFilesToEncrpt);
    if not FUseStream and (Length(FFileName) = 0) then
      raise EDefaceError.Create(Format(SErrFilenameFailed, ['']));

    if Length(PassWord) = 0 then PassWord := 'Empty';
    FEncStream.SetSize(0);
    FTmpStream.SetSize(0);

    if (not ZipFiles) or (FTmpStream.Size = 0) then         //Zuerst alle Files packen
      raise EDefaceError.Create(SErrZipFailed);

    if (not EncryptZip) or (FEncStream.Size = 0) then       //Jetzt ZipStream verschlüsseln
      raise EDefaceError.Create(SErrEncryptFailed);

    EncodeFile;                                             //Jetzt verschlüsseltes File codieren
    if FEncStream.Size = 0 then
      raise EDefaceError.Create(SErrInDefaceFailed);

    if not FUseStream then SaveDefaceFile;
  except
    on e: Exception do
      ShowMessage(e.Message);
  end;
  Progress.Finish;
end;

procedure TCustomDefacer.OutDeface;                         //Entschlüsselung
begin
  try
    if (FState = dsEmpty) and not FUseStream then
      LoadFromFile(FFileName);

    if (FState <> dsCoded) or (FEncStream.Size = 0) then
      raise EDefaceError.Create(SErrNoEncryptedData);

    if Length(PassWord) = 0 then PassWord := 'Empty';

    DecodeFile;                                             //verschlüsseltes File decodieren

    if not DecryptZip then                                  //Jetzt erstellen ZipStream
      raise EDefaceError.Create(SErrDecryptFailed);

    if not ReadZipFileNames then
      raise EDefaceError.Create(SErrReadZipNames);

    if not UnZipFiles then
      raise EDefaceError.Create(SErrUnZipFailed);
  except
    on e: Exception do
      ShowMessage(e.Message);
  end;
  Progress.Finish;
end;

procedure TCustomDefacer.LoadFromFile(aFileName: String);   //Laden vom File
begin
  FFileName := aFileName;
  FEncStream.SetSize(0);
  FTmpStream.SetSize(0);
  FEncStream.LoadFromFile(FFileName);
  FState := dsCoded;
end;

procedure TCustomDefacer.LoadFromStream(Stream: TStream);   //Laden vom Stream
begin
  FEncStream.SetSize(0);
  FTmpStream.SetSize(0);
  FEncStream.LoadFromStream(Stream);
  FState := dsCoded;
end;

procedure TCustomDefacer.SaveToFile(aFileName: String);     //Speichern in Datei
begin
  FFileName     := aFileName;
  FUseStream := False;
  InDeface;
end;

procedure TCustomDefacer.SaveToStream(Stream: TStream);     //Speichern in Stream
begin
  FUseStream := True;
  InDeface;
  if FEncStream.Size > 0 then
    FEncStream.SaveToStream(Stream);
end;

procedure TCustomDefacer.SaveToDirectory(Dir: String);      //Speichern der Dateien in Verzeichnis
begin
  FOutputDir := Dir;
  OutDeface;
end;


constructor TCustomDefacer.Create;                          //Constructor
begin
  inherited Create;
  FCodeData            := TCodeData.Create;
  FCodeFileList        := TStringList.Create;
  FPackFileList        := TStringList.Create;
  FEncStream           := TMemoryStream.Create;
  FTmpStream           := TMemoryStream.Create;
  FPackFileList.Sorted := True;
  Clear;
end;

destructor TCustomDefacer.Destroy;                          //Destructor
begin
  FreeAndNil(FTmpStream);
  FreeAndNil(FEncStream);
  FreeAndNil(FPackFileList);
  FreeAndNil(FCodeFileList);
  FreeAndNil(FCodeData);
  inherited Destroy;
end;

procedure TCustomDefacer.Clear;                             //Clear
begin
  FEncStream.SetSize(0);
  FTmpStream.SetSize(0);
  FFileName     := '';
  FPassWord     := '';
  FOutputDir    := '';
  FState        := dsEmpty;
  FUseStream := False;
end;

procedure TCustomDefacer.CodeFileAdd(                       //Fügt ein File zur CodeFileList hinzu
  const aFileName: String);
begin
  FCodeFileList.Add(aFileName);
end;

procedure TCustomDefacer.CodeFileDel(aIndex: Integer);      //Löscht File an Index...
begin
  if (aIndex >= 0) and (aIndex < FCodeFileList.Count) then
    FCodeFileList.Delete(aIndex);
end;

{ TInDefacer }

procedure TInDefacer.InDeface;
begin
  inherited InDeface;
end;

procedure TInDefacer.ProgressStart;
var
  i: Integer;
begin
  inherited ProgressStart;
  Progress.Add(SReadFileStructure,    1, 0,    15);
  Progress.Add(SCreateFileStructur,  10, 0,    15);
  Progress.Add(SCompressFiles,      500, 0, 45444);
  Progress.Add(SEncryptArchive,       1, 0,  6115);
  Progress.Add(SCreateEncrArchive,    1, 0,    15);

  for i := 0 to FCodeFileList.Count - 1 do
    Progress.Add(Format(SEnCodeWithFile, [FCodeFileList[i]]), 100, 0, 16800);

  Progress.Add(SSaveArchive,          1, 0,  1750);
  Progress.Start;
end;

procedure TInDefacer.SaveToFile(aFileName: String);
begin
  inherited SaveToFile(aFileName);
end;

procedure TInDefacer.SaveToStream(Stream: TStream);
begin
  inherited SaveToStream(Stream);
end;

{ TOutDefacer }

procedure TOutDefacer.ProgressStart;
var
  i: Integer;
begin
  inherited ProgressStart;
  Progress.Add(SLoadArchive,         1, 0,   187);

  for i := FCodeFileList.Count - 1 downto 0 do
    Progress.Add(Format(SDeCodeWithFile, [FCodeFileList[i]]), 100, 0, 17000);

  Progress.Add(SDecryptArchive,      1, 0,  5242);
  Progress.Add(SCreateZipArchive,    1, 0,  1108);
  Progress.Add(SReadFileStructure, 100, 0,    31);
  Progress.Add(SDeCompressFiles,   500, 0, 10608);

  Progress.Start;
end;

procedure TOutDefacer.OutDeface;
begin
  inherited OutDeface;
end;

procedure TOutDefacer.LoadFromFile(aFileName: String);
begin
  inherited LoadFromFile(aFileName);
end;

procedure TOutDefacer.LoadFromStream(Stream: TStream);
begin
  inherited LoadFromStream(Stream);
end;

procedure TOutDefacer.SaveToDirectory(Dir: String);
begin
  inherited SaveToDirectory(Dir);
end;

end.

