{*******************************************************************************
  Copyright (C) Michael W. Vogel m-w-vogel@gmx.de

  The contents of this file are subject to the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 3.0 of the License,
  or (at your option) any later version.
  You may obtain a copy of the LGPL at http://www.gnu.org/copyleft/.

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
  specific language governing rights and limitations under the License.

  Created 02.09.2015
*******************************************************************************}

unit VirtualFileTrees;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, VirtualTrees, Controls, ImgList,
  LCLType, LResources, FileUtil, PointerLists,
  ArchiveParser;

type

  EFileTreeError = class(Exception);
  PFileTreeData = ^TFileTreeData;
  TFileTreeData = record
    PathID: Integer;
    FileName: String;
    IsDirectory: Boolean;
    IsClarified: Boolean;
  end;

  TPathNode = record
    Path: String;
    Node: PVirtualNode;
  end;

  TVirtualFileTree = class;
  TFTOnReplaceNode = procedure(Node: PVirtualNode; FileName: String; var Replace: Boolean) of object;

  TFileOption = (foAllowDeleteNode, foObservePath);
  TFileOptions = set of TFileOption;

const
  DefaultFileOptions = [];

type

  { TFileTreeOptions }

  TFileTreeOptions = class(TStringTreeOptions)
  private
    FFileOptions: TFileOptions;
  published
    property FileOptions: TFileOptions read FFileOptions
      write FFileOptions default DefaultFileOptions;

    property AnimationOptions;
    property AutoOptions;
    property ExportMode;
    property MiscOptions;
    property PaintOptions;
    property SelectionOptions;
    property StringOptions;
  end;

  { TTreeSearcher }

  TTreeSearcher = class(TFileSearcher)
  private
    FAddDirs: Boolean;
    FDirList: TPointerList;
    FTree: TVirtualFileTree;
    FParentNode: PVirtualNode;
    FPathID: Integer;
    function FindNode(aDir: String): PVirtualNode;
  protected
    procedure DoFileFound; override;
    procedure DoDirectoryFound; override;
  public
    procedure Search(ASearchPath: String;
      ASearchMask: String = '';
      WithSubDir: Boolean = False);
    constructor Create(ATree: TVirtualFileTree; ParentNode: PVirtualNode);
    destructor Destroy; override;
  end;

  { TCustomVirtualFileTree }

  TCustomVirtualFileTree = class(TCustomVirtualStringTree)
  private
    FOnFreeFileNode: TVTFreeNodeEvent;
    FOnReplaceNode: TFTOnReplaceNode;
    FPaths: array of TPathNode;
    FImages: TImageList;
    function GetOptions: TFileTreeOptions;
    procedure SetOptions(const Value: TFileTreeOptions);
    function AddPathID(const FilePath: String): Integer;
    procedure AddDirPhysical(const Dir: String; ParentNode: PVirtualNode = nil);
    procedure AddDirMemory(const Dir: String);
    procedure AddFilePhysical(const FileName: String; ParentNode: PVirtualNode = nil);
    procedure AddFileMemory(const FileName: String);
    function ExtractDirectory(const FileName: String): String;
  protected
    function AddNode(FileName: String; PathID: Integer; IsDir: Boolean; aParent: PVirtualNode = nil): PVirtualNode;
    procedure AddPhysicalDirectory(ParentNode: PVirtualNode;
      const Dir: String; WithSubDir: Boolean);
    function DoCompare(Node1, Node2: PVirtualNode; Column: TColumnIndex): Integer; override;
    procedure DoExpanded(Node: PVirtualNode); override;
    procedure DoFreeFileNode({%H-}Sender: TBaseVirtualTree;
      Node: PVirtualNode);
    function DoGetImageIndex(Node: PVirtualNode; Kind: TVTImageKind; Column: TColumnIndex;
      var Ghosted: Boolean; var Index: Integer): TCustomImageList; override;
    procedure DoGetText(Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType;
      var aText: String); override;
    function DoKeyAction(var CharCode: Word; var Shift: TShiftState): Boolean; override;
    function FindNode(aParent: PVirtualNode; FileName: String): PVirtualNode;
    function GetOptionsClass: TTreeOptionsClass; override;
    procedure ValidateNodeDataSize(var Size: Integer); override;
    property Options: TFileTreeOptions read GetOptions write SetOptions;
    property OnFreeFileNode: TVTFreeNodeEvent read FOnFreeFileNode write FOnFreeFileNode;
    property OnReplaceNode: TFTOnReplaceNode read FOnReplaceNode write FOnReplaceNode;
  public
    constructor Create(AOwner: TComponent); override;
    procedure Clear; override;
    procedure AddDir(const Dir: String; ParentNode: PVirtualNode = nil);
    procedure AddFile(const FileName: String; ParentNode: PVirtualNode = nil);
    function GetPathID(FilePath: String): Integer;
    function GetPath(FilePathID: Integer): String;
    function GetNode(FilePathID: Integer): PVirtualNode;
    function GetFileName(Node: PVirtualNode): String; overload;
    function GetFileName(Index: Cardinal): String; overload;
    procedure GetFilesOrDirectories(FileList: TStringList;
      WithPath, GetFiles, GetDirs: Boolean; ForParsing: Boolean = False);
    procedure GetFiles(FileList: TStringList; WithPath: Boolean);
    procedure GetDirectories(FileList: TStringList; WithPath: Boolean);
  end;

  TVirtualFileTree = class(TCustomVirtualFileTree)
  public
    property Canvas;
    property RangeX;
    property LastDragEffect;
  published
    {$ifdef EnableAccessible}
    property AccessibleName;
    {$endif}
    property Action;
    property Align;
    property Alignment;
    property Anchors;
    property AnimationDuration;
    property AutoExpandDelay;
    property AutoScrollDelay;
    property AutoScrollInterval;
    property Background;
    property BackgroundOffsetX;
    property BackgroundOffsetY;
    property BiDiMode;
    property BorderSpacing;
    property BorderStyle default bsSingle;
    property BottomSpace;
    property ButtonFillMode;
    property ButtonStyle;
    property BorderWidth;
    property ChangeDelay;
    property CheckImageKind;
    property ClipboardFormats;
    property Color;
    property Colors;
    property Constraints;
    property CustomCheckImages;
    property DefaultNodeHeight;
    property DefaultPasteMode;
    property DefaultText;
    property DragCursor;
    property DragHeight;
    property DragKind;
    property DragImageKind;
    property DragMode;
    property DragOperations;
    property DragType;
    property DragWidth;
    property DrawSelectionMode;
    property EditDelay;
    property EmptyListMessage;
    property Enabled;
    property Font;
    property Header;
    property HintMode;
    property HotCursor;
    property Images;
    property IncrementalSearch;
    property IncrementalSearchDirection;
    property IncrementalSearchStart;
    property IncrementalSearchTimeout;
    property Indent;
    property LineMode;
    property LineStyle;
    property Margin;
    property NodeAlignment;
    property NodeDataSize;
    property OperationCanceled;
    property Options;
    property ParentBiDiMode;
    property ParentColor default False;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property RootNodeCount;
    property ScrollBarOptions;
    property SelectionBlendFactor;
    property SelectionCurveRadius;
    property ShowHint;
    property StateImages;
    property TabOrder;
    property TabStop default True;
    property TextMargin;
    property Visible;
    property WantTabs;

    property OnAddToSelection;
    property OnAdvancedHeaderDraw;
    property OnAfterAutoFitColumn;
    property OnAfterAutoFitColumns;
    property OnAfterCellPaint;
    property OnAfterColumnExport;
    property OnAfterColumnWidthTracking;
    property OnAfterGetMaxColumnWidth;
    property OnAfterHeaderExport;
    property OnAfterHeaderHeightTracking;
    property OnAfterItemErase;
    property OnAfterItemPaint;
    property OnAfterNodeExport;
    property OnAfterPaint;
    property OnAfterTreeExport;
    property OnBeforeAutoFitColumn;
    property OnBeforeAutoFitColumns;
    property OnBeforeCellPaint;
    property OnBeforeColumnExport;
    property OnBeforeColumnWidthTracking;
    property OnBeforeDrawTreeLine;
    property OnBeforeGetMaxColumnWidth;
    property OnBeforeHeaderExport;
    property OnBeforeHeaderHeightTracking;
    property OnBeforeItemErase;
    property OnBeforeItemPaint;
    property OnBeforeNodeExport;
    property OnBeforePaint;
    property OnBeforeTreeExport;
    property OnCanSplitterResizeColumn;
    property OnCanSplitterResizeHeader;
    property OnCanSplitterResizeNode;
    property OnChange;
    property OnChecked;
    property OnChecking;
    property OnClick;
    property OnCollapsed;
    property OnCollapsing;
    property OnColumnClick;
    property OnColumnDblClick;
    property OnColumnExport;
    property OnColumnResize;
    property OnColumnWidthDblClickResize;
    property OnColumnWidthTracking;
    property OnCompareNodes;
    property OnContextPopup;
    property OnCreateDataObject;
    property OnCreateDragManager;
    property OnCreateEditor;
    property OnDblClick;
    property OnDragAllowed;
    property OnDragOver;
    property OnDragDrop;
    property OnDrawHint;
    property OnDrawText;
    property OnEditCancelled;
    property OnEdited;
    property OnEditing;
    property OnEndDock;
    property OnEndDrag;
    property OnEndOperation;
    property OnEnter;
    property OnExit;
    property OnExpanded;
    property OnExpanding;
    property OnFocusChanged;
    property OnFocusChanging;
    property OnFreeFileNode;
    property OnGetCellIsEmpty;
    property OnGetCursor;
    property OnGetHeaderCursor;
    property OnGetText;
    property OnPaintText;
    property OnGetHelpContext;
    property OnGetHintKind;
    property OnGetHintSize;
    property OnGetImageIndex;
    property OnGetImageIndexEx;
    property OnGetImageText;
    property OnGetHint;
    property OnGetLineStyle;
    property OnGetNodeDataSize;
    property OnGetPopupMenu;
    property OnGetUserClipboardFormats;
    property OnHeaderClick;
    property OnHeaderDblClick;
    property OnHeaderDragged;
    property OnHeaderDraggedOut;
    property OnHeaderDragging;
    property OnHeaderDraw;
    property OnHeaderDrawQueryElements;
    property OnHeaderHeightDblClickResize;
    property OnHeaderHeightTracking;
    property OnHeaderMouseDown;
    property OnHeaderMouseMove;
    property OnHeaderMouseUp;
    property OnHotChange;
    property OnIncrementalSearch;
    property OnInitChildren;
    property OnInitNode;
    property OnKeyAction;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnLoadNode;
    property OnLoadTree;
    property OnMeasureItem;
    property OnMeasureTextWidth;
    property OnMeasureTextHeight;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseWheel;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnNewText;
    property OnNodeClick;
    property OnNodeCopied;
    property OnNodeCopying;
    property OnNodeDblClick;
    property OnNodeExport;
    property OnNodeHeightDblClickResize;
    property OnNodeHeightTracking;
    property OnNodeMoved;
    property OnNodeMoving;
    property OnPaintBackground;
    property OnReplaceNode;
    property OnRemoveFromSelection;
    property OnRenderOLEData;
    property OnResetNode;
    property OnResize;
    property OnSaveNode;
    property OnSaveTree;
    property OnScroll;
    property OnShortenString;
    property OnShowScrollbar;
    property OnStartDock;
    property OnStartDrag;
    property OnStartOperation;
    property OnStateChange;
    property OnStructureChange;
    property OnUpdating;
    property OnUTF8KeyPress;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Virtual Controls', [TVirtualFileTree]);
end;

{ TTreeSearcher }

function TTreeSearcher.FindNode(aDir: String): PVirtualNode;
var
  aNode: PVirtualNode;
  Data: PFileTreeData;
begin
  Result := nil;
  aNode  := FDirList.First;
  while Assigned(aNode) do begin
    Data := FTree.GetNodeData(aNode);
    if Assigned(Data) and (Data^.FileName = aDir) then begin
      Result := aNode;
      Exit;
    end;
    aNode := FDirList.Next;
  end;
end;

procedure TTreeSearcher.DoFileFound;
begin
  FTree.AddNode(StringReplace(FileName, Path, '', []), FPathID, False, FParentNode);
end;

procedure TTreeSearcher.DoDirectoryFound;
begin
  if FAddDirs then
    FDirList.Add(FTree.AddNode(StringReplace(FileName, Path, '', []), FPathID, True, FParentNode))
  else
    FTree.AddNode(StringReplace(FileName, Path, '', []), FPathID, True, FParentNode);
end;

procedure TTreeSearcher.Search(ASearchPath: String;
  ASearchMask: String; WithSubDir: Boolean);
begin
  FAddDirs := WithSubDir;
  FPathID  := FTree.GetPathID(ASearchPath);
  if FPathID < 0 then FPathID := FTree.AddPathID(ASearchPath);
  inherited Search(ASearchPath, ASearchMask, False);
end;

constructor TTreeSearcher.Create(ATree: TVirtualFileTree;
  ParentNode: PVirtualNode);
begin
  inherited Create;
  FParentNode := ParentNode;
  FAddDirs    := True;
  FDirList    := TPointerList.Create;
  FTree       := ATree;
end;

destructor TTreeSearcher.Destroy;
begin
  FreeAndNil(FDirList);
  inherited Destroy;
end;

{ TCustomVirtualFileTree }

function TCustomVirtualFileTree.GetOptions: TFileTreeOptions;
begin
  Result := TreeOptions as TFileTreeOptions;
end;

procedure TCustomVirtualFileTree.SetOptions(
  const Value: TFileTreeOptions);
begin
  TreeOptions := Value;
end;

function TCustomVirtualFileTree.AddPathID(
  const FilePath: String): Integer;
begin
  SetLength(FPaths, Length(FPaths) + 1);
  Result := High(FPaths);
  FPaths[Result].Path := FilePath;
  FPaths[Result].Node := nil;
end;

procedure TCustomVirtualFileTree.AddDirPhysical(
  const Dir: String; ParentNode: PVirtualNode);
var
  Node: PVirtualNode;
  Data: PFileTreeData;
  aPath, aDirName: String;
  aPathID: Integer;
begin
  BeginUpdate;
  aDirName := ExtractFileName(Dir);
  aPath := ExtractDirectory(Dir);
  aPathID := GetPathID(aPath);
  if aPathID < 0 then begin
    aPathID := AddPathID(aPath);
    Node := AddNode(aDirName, aPathID, True, ParentNode);
    FPaths[aPathID].Node := Node;
  end else
    Node := AddNode(aDirName, aPathID, True, ParentNode);
  Data := GetNodeData(Node);
  if Assigned(Data) then
    Data^.IsClarified := True;
  AddPhysicalDirectory(Node, Dir, True);
  EndUpdate;
end;

procedure TCustomVirtualFileTree.AddDirMemory(
  const Dir: String);
var
  aParent: PVirtualNode;
  Node: PVirtualNode;
  Data: PFileTreeData;
  ParentDirectory: String;
  aDirName: String;
  aPathID: Integer;
begin
  if (Length(Dir) = 0) or (GetPathID(Dir) >= 0) then Exit;
  if Dir[Length(Dir)] = DirectorySeparator then
    raise EFileTreeError.Create('Directory should end with no DirectorySeparator');

  aDirName        := ExtractFileName(Dir);
  ParentDirectory := ExtractDirectory(Dir);
  aPathID         := GetPathID(ParentDirectory);
  if (aPathID < 0) and (Length(ParentDirectory) > 0) then begin
    AddDirMemory(ParentDirectory);
    aPathID := GetPathID(ParentDirectory);
  end;

  BeginUpdate;
  if aPathID < 0 then
    aParent := nil
  else
    aParent := FPaths[aPathID].Node;

  aPathID := AddPathID(Dir);
  Node := AddNode(aDirName, aPathID, True, aParent);
  FPaths[High(FPaths)].Node := Node;
  Data := GetNodeData(Node);
  if Assigned(Data) then
    Data^.IsClarified := True;
  EndUpdate;
end;

procedure TCustomVirtualFileTree.AddFilePhysical(
  const FileName: String; ParentNode: PVirtualNode);
var
  aPath, aFileName: String;
  aPathID: Integer;
begin
  BeginUpdate;
  aFileName := ExtractFileName(FileName);
  aPath := ExtractDirectory(FileName);
  aPathID := GetPathID(aPath);
  if aPathID < 0 then begin
    aPathID := AddPathID(aPath);
    FPaths[aPathID].Node := AddNode(aFileName, aPathID, False, ParentNode);
  end else
    AddNode(aFileName, aPathID, False, ParentNode);
  EndUpdate;
end;

procedure TCustomVirtualFileTree.AddFileMemory(
  const FileName: String);
var
  aParent: PVirtualNode;
  Node: PVirtualNode;
  Data: PFileTreeData;
  ParentDirectory: String;
  aFileName: String;
  aPathID: Integer;
begin
  if (Length(FileName) = 0) then Exit;

  aFileName       := ExtractFileName(FileName);
  ParentDirectory := ExtractDirectory(FileName);
  aPathID         := GetPathID(ParentDirectory);
  if (aPathID < 0) and (Length(ParentDirectory) > 0) then begin
    AddDirMemory(ParentDirectory);
    aPathID := GetPathID(ParentDirectory);
  end;

  BeginUpdate;
  if aPathID < 0 then
    aParent := nil
  else
    aParent := FPaths[aPathID].Node;

  Node := AddNode(aFileName, aPathID, False, aParent);
  Data := GetNodeData(Node);
  if Assigned(Data) then
    Data^.IsClarified := True;
  EndUpdate;
end;

function TCustomVirtualFileTree.ExtractDirectory(
  const FileName: String): String;
begin
  Result := ExtractFileDir(FileName);
  if (Length(Result) > 0) and (Result[Length(Result)] = DirectorySeparator) then
    SetLength(Result, Length(Result) - 1);
end;

function TCustomVirtualFileTree.AddNode(FileName: String;
  PathID: Integer; IsDir: Boolean; aParent: PVirtualNode
  ): PVirtualNode;
var
  Data: PFileTreeData;
  ReplaceOldNode: Boolean;
begin
  Result := FindNode(aParent, FileName);
  if Assigned(Result) then begin
    ReplaceOldNode := True;
    if Assigned(FOnReplaceNode) then
      FOnReplaceNode(Result, GetPath(PathID) + FileName, ReplaceOldNode);
    if not ReplaceOldNode then Exit;
    DeleteNode(Result);
  end;
  Result := AddChild(aParent);
  Data   := GetNodeData(Result);
  if not Assigned(Data) then Exit;
  Data^.IsDirectory := IsDir;
  Data^.IsClarified := not IsDir;
  Data^.FileName    := FileName;
  Data^.PathID      := PathID;
end;

procedure TCustomVirtualFileTree.AddPhysicalDirectory(
  ParentNode: PVirtualNode; const Dir: String;
  WithSubDir: Boolean);
var
  FS: TTreeSearcher;
begin
  FS := TTreeSearcher.Create(TVirtualFileTree(Self), ParentNode);
  try
    FS.Search(Dir, AllFilesMask, WithSubDir);
  finally
    FreeAndNil(FS);
  end;
end;

function TCustomVirtualFileTree.DoCompare(Node1,
  Node2: PVirtualNode; Column: TColumnIndex): Integer;
var
  Data1, Data2: PFileTreeData;
begin
  Result := 0;
  if Assigned(OnCompareNodes) then begin
    OnCompareNodes(Self, Node1, Node2, Column, Result);
    Exit;
  end;

  Data1 := GetNodeData(Node1);
  if not Assigned(Data1) then Exit;
  Data2 := GetNodeData(Node2);
  if not Assigned(Data2) then Exit;
  if Data1^.IsDirectory and not Data2^.IsDirectory then begin
    Result:=-1;
    Exit;
  end;
  if Data2^.IsDirectory and not Data1^.IsDirectory then begin
    Result:=1;
    Exit;
  end;
  Result := CompareStr(Data1^.FileName, Data2^.FileName);
end;

procedure TCustomVirtualFileTree.DoExpanded(Node: PVirtualNode);
var
  aNode: PVirtualNode;
  Data: PFileTreeData;
begin
  Data := GetNodeData(Node);
  if Assigned(Data) then
    Data^.IsClarified := True;

  aNode := Node^.FirstChild;
  while Assigned(aNode) do begin
    Data := GetNodeData(aNode);
    if Assigned(Data) and Data^.IsDirectory and not Data^.IsClarified then
      AddPhysicalDirectory(aNode, GetFileName(aNode), False);
    Data^.IsClarified := True;
    aNode := aNode^.NextSibling;
  end;

  inherited DoExpanded(Node);
end;

procedure TCustomVirtualFileTree.DoFreeFileNode(
  Sender: TBaseVirtualTree; Node: PVirtualNode);
var
  Data: PFileTreeData;
begin
  if Assigned(FOnFreeFileNode) then
    FOnFreeFileNode(Self, Node);

  Data := GetNodeData(Node);
  if not Assigned(Data) then Exit;
  Data^.FileName := '';
end;

function TCustomVirtualFileTree.DoGetImageIndex(
  Node: PVirtualNode; Kind: TVTImageKind;
  Column: TColumnIndex; var Ghosted: Boolean;
  var Index: Integer): TCustomImageList;
var
  Data: PFileTreeData;
begin
  Data := GetNodeData(Node);
  if Assigned(Data) and Assigned(Images) then
    if not Data^.IsDirectory then
      Index := 0
    else
      if vsExpanded in Node^.States then
        Index := 2
      else
        Index := 1;

  Result := inherited DoGetImageIndex(Node, Kind, Column,
    Ghosted, Index);
end;

procedure TCustomVirtualFileTree.DoGetText(Node: PVirtualNode;
  Column: TColumnIndex; TextType: TVSTTextType;
  var aText: String);
var
  Data: PFileTreeData;
begin
  Data := GetNodeData(Node);
  if Assigned(Data) then
    aText := Data^.FileName;
  inherited DoGetText(Node, Column, TextType, aText);
end;

function TCustomVirtualFileTree.DoKeyAction(var CharCode: Word;
  var Shift: TShiftState): Boolean;
begin
  Result := inherited DoKeyAction(CharCode, Shift);
  if not Result then Exit;

  if (CharCode = VK_DELETE) and (foAllowDeleteNode in Options.FileOptions) then
    DeleteSelectedNodes;
end;

function TCustomVirtualFileTree.FindNode(
  aParent: PVirtualNode; FileName: String): PVirtualNode;
var
  aNode: PVirtualNode;
  Data: PFileTreeData;
begin
  Result := nil;
  if Assigned(aParent) then
    aNode := aParent^.FirstChild
  else
    aNode := GetFirst(False);

  while Assigned(aNode) do begin
    Data := GetNodeData(aNode);
    if Assigned(Data) and (Data^.FileName = FileName) then begin
      Result := aNode;
      Exit;
    end;
    aNode := GetNextSibling(aNode);
  end;
end;

function TCustomVirtualFileTree.GetOptionsClass: TTreeOptionsClass;
begin
  Result := TFileTreeOptions;
end;

procedure TCustomVirtualFileTree.ValidateNodeDataSize(
  var Size: Integer);
begin
  Size := SizeOf(TFileTreeData);
  if Assigned(OnGetNodeDataSize) then
    OnGetNodeDataSize(Self, Size);
end;


constructor TCustomVirtualFileTree.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetLength(FPaths, 0);
  OnFreeNode := @DoFreeFileNode;
  FImages := TImageList.Create(Self);
  FImages.AddLazarusResource('file');
  FImages.AddLazarusResource('directory');
  FImages.AddLazarusResource('opendirectory');
  Images := FImages;
end;

procedure TCustomVirtualFileTree.Clear;
begin
  inherited Clear;
  SetLength(FPaths, 0);
end;

procedure TCustomVirtualFileTree.AddDir(const Dir: String;
  ParentNode: PVirtualNode);
begin
  if foObservePath in Options.FileOptions then
    AddDirMemory(Dir)
  else
    AddDirPhysical(Dir, ParentNode);
end;

procedure TCustomVirtualFileTree.AddFile(
  const FileName: String; ParentNode: PVirtualNode);
begin
  if foObservePath in Options.FileOptions then
    AddFileMemory(FileName)
  else
    AddFilePhysical(FileName, ParentNode);
end;

function TCustomVirtualFileTree.GetPathID(FilePath: String
  ): Integer;
var
  i: Integer;
begin
  Result := -1;
  if Length(FilePath) = 0 then Exit;
  if FilePath[Length(FilePath)] = DirectorySeparator then
    raise EFileTreeError.Create('FilePath should not end with DirectorySeparator');

  for i := 0 to High(FPaths) do
    if FilePath = FPaths[i].Path then begin
      Result := i;
      Exit;
    end;
end;

function TCustomVirtualFileTree.GetPath(FilePathID: Integer
  ): String;
begin
  if (FilePathID >= 0) and (FilePathID < Length(FPaths)) then
    Result := FPaths[FilePathID].Path
  else
    Result := '';
end;

function TCustomVirtualFileTree.GetNode(FilePathID: Integer
  ): PVirtualNode;
begin
  if (FilePathID >= 0) and (FilePathID < Length(FPaths)) then
    Result := FPaths[FilePathID].Node
  else
    Result := nil;
end;

function TCustomVirtualFileTree.GetFileName(
  Node: PVirtualNode): String;
var
  Data: PFileTreeData;
begin
  if not Assigned(Node) then begin
    Result := '';
    Exit;
  end;

  Data := GetNodeData(Node);
  if not Assigned(Data) then begin
    Result := '';
    Exit;
  end;

  Result := GetPath(Data^.PathID) + DirectorySeparator + Data^.FileName;
end;

function TCustomVirtualFileTree.GetFileName(Index: Cardinal
  ): String;
var
  Node: PVirtualNode;
begin
  if Index >= TotalCount then
  begin
    Result := '';
    Exit;
  end;

  Node := GetFirst(False);

  while Assigned(Node) and (Index > 0) do begin
    Node := GetNext(Node);
    Dec(Index);
  end;

  if Assigned(Node) and (Index = 0) then
    Result := GetFileName(Node);
end;

procedure TCustomVirtualFileTree.GetFilesOrDirectories(
  FileList: TStringList; WithPath, GetFiles,
  GetDirs: Boolean; ForParsing: Boolean);

var
  ParentPath: String;

  function WithOutPath(const FileName: String): String;
  begin
    if Pos(ParentPath, FileName) = 1 then
      Result := Copy(FileName, Length(ParentPath) + 2, Length(FileName))
    else
      Result := FileName;
  end;
  procedure DirectoryAdd(aDir: String);
  var
    SL: TStringList;
    i: Integer;
  begin
    SL := TStringList.Create;

    if GetDirs then begin
      FindAllDirectories(SL, aDir);
      for i := 0 to SL.Count - 1 do
        if ForParsing then
          FileList.Add(
            PackFileName(True, SL[i], WithOutPath(SL[i])))
        else
          if WithPath then
            FileList.Add(SL[i])
          else
            FileList.Add(WithOutPath(SL[i]));
    end;

    if GetFiles then begin
      FindAllFiles(SL, aDir, '', True, faAnyFile);
      for i := 0 to SL.Count - 1 do
        if ForParsing then
          FileList.Add(
            PackFileName(False, SL[i], WithOutPath(SL[i])))
        else
          if WithPath then
            FileList.Add(SL[i])
          else
            FileList.Add(WithOutPath(SL[i]));
    end;

    FreeAndNil(SL);
  end;

var
  aNode: PVirtualNode;
  Data: PFileTreeData;
begin
  FileList.Clear;
  aNode := GetFirst(False);
  ParentPath := '';

  while Assigned(aNode) do begin
    Data := GetNodeData(aNode);
    if Assigned(Data) then

      if aNode^.Parent = RootNode then
        ParentPath := GetPath(Data^.PathID);

      if not Data^.IsDirectory then begin
        if GetFiles then
          if ForParsing then
            FileList.Add(
              PackFileName(True, GetFileName(aNode), WithOutPath(GetFileName(aNode))))
          else
            if WithPath then
              FileList.Add(GetFileName(aNode))
            else
              FileList.Add(WithOutPath(GetFileName(aNode)));
      end else begin
        if GetDirs then
          if ForParsing then
            FileList.Add(
              PackFileName(False, GetFileName(aNode), WithOutPath(GetFileName(aNode))))
          else
            if WithPath then
              FileList.Add(GetFileName(aNode))
            else
              FileList.Add(WithOutPath(GetFileName(aNode)));
        if not Data^.IsClarified then
          DirectoryAdd(GetFileName(aNode));
      end;
    aNode := GetNext(aNode, True);
  end;
end;

procedure TCustomVirtualFileTree.GetFiles(
  FileList: TStringList; WithPath: Boolean);
begin
  GetFilesOrDirectories(FileList, WithPath, True, False);
end;

procedure TCustomVirtualFileTree.GetDirectories(
  FileList: TStringList; WithPath: Boolean);
begin
  GetFilesOrDirectories(FileList, WithPath, False, True);
end;

initialization
  {$I treeimages.lrs}
  {$I virtualfiletree_package.lrs}

end.

